rm(list=ls())
library(skellam)
#National Teams or Premier League?
NationalTeams=F
competition<-ifelse(NationalTeams==T, "NationalTeams","PremierLeague")
load(paste0("results",competition,".Rdata"))

#source Likelihoodfunctions
source("mlFunctions.R")

#function to look op the initial values
doubleListLookup <- function(myList,first,second,defaultLength){
  out <- rep(0,defaultLength)
  if(!is.null(myList) && !is.null(myList[[as.character(first)]][[as.character(second)]])){
    out <- myList[[as.character(first)]][[as.character(second)]]
  }
  
  out
}


computePredictions <- function(startPredictionDate,predictionPeriodYears,
                                   halfPeriods,targetSaveFile,
                                   continueAnalysis,NationalTeams=T,
                                   methods = c("BTD","BTDgd","BTT","BTTgd","TM","TMgd","P1","P2","KN1","KN2")){
  summary <- NULL
  nbHalfPeriods <- length(halfPeriods)
  results <- results[order(results$date),]
  if(NationalTeams==F){
  results$impScore <- 1
  }
  # calculations to speed up estimation
  results$mingoal<-pmin(results$home_score, results$away_score)
  xymin<-max(results$mingoal)
  results$kn1<-(results$mingoal>0)*results$home_score*results$away_score
  results$kn2<-(results$mingoal>1)*results$kn1*(results$home_score-1)*(results$away_score-1)/2
  results$kn3<-(results$mingoal>2)*results$kn2*(results$home_score-2)*(results$away_score-2)/6
  results$kn4<-(results$mingoal>3)*results$kn3*(results$home_score-3)*(results$away_score-3)/24
  results$kn5<-(results$mingoal>4)*results$kn4*(results$home_score-4)*(results$away_score-4)/120
  results$gdweights<-1
  results$gdweights[results$goaldif!=0]<-log2(abs(results$goaldif[results$goaldif!=0])+1)
  
  # Preprocess data
  teams <- unique(sort(c(results$home_team,results$away_team)))
  nb.teams <- length(teams)
  results$homeId <- match(results$home_team,teams)
  results$visitId <- match(results$away_team,teams)
  
  # Use saved summary as a starting point if continueAnalysis is true and
  # there are already processed matches in summary
  successfulLoad <- FALSE
  try({
    load(targetSaveFile)
    successfulLoad <- TRUE
  },silent=T)
  if(continueAnalysis && successfulLoad){
    startPredictDate <- max(summary$date)+1
  } else{
    summary <- NULL
    startPredictDate <- startPredictionDate
  }
  # Keep track of the iterations in order to save each saveEachLoops loops
  iteration <- 1
  # Initial values of the ML estimation 
  inits <<- NULL
  nDatarow <- nrow(results)
  
  while(TRUE){
    # Calculate the maximum date range that does not play the same teams twice.
    startIndex <- which(results$date>=startPredictDate)[1]
    
    # Exit the infite loop when all matches have been processed
    if(is.na(startIndex)){
      break
    }
    intermediateNextDate <- results$date[startIndex]
    endPredictDate <- results$date[startIndex]
    playedTeams <- c(results$home_team[startIndex],results$away_team[startIndex])
    nextIndex <- startIndex+1
    
    # Split data from fixture to fixture and calculate predictions for future
    # matches. A fixture in the national team setting is defined as the maximum
    # date range that does not have the same team play twice.
    while(TRUE){
      if(nextIndex>nDatarow){
        break
      }
      if(any(c(results$home_team[nextIndex],results$away_team[nextIndex]) %in%
             playedTeams)){
        break
      } else{
        playedTeams <- c(playedTeams,
                         c(results$home_team[nextIndex],
                           results$away_team[nextIndex]))
        nextIndex <- nextIndex+1
        if(nextIndex>nDatarow){
          break
        }
        if(results$date[nextIndex]>intermediateNextDate){
          endPredictDate <- intermediateNextDate
          intermediateNextDate <- results$date[nextIndex-1]
        }
      }
    }
    
    # Display the processed date range and predicted matches count
    cat(iteration," - ","Processing date range ",
        as.character(startPredictDate)," - ",as.character(endPredictDate)," (",
        sum(results$date>=startPredictDate & results$date<=endPredictDate),
        ")","\n",sep="")
    # Calculate the modeling date range
    startModelDate <- as.POSIXlt(startPredictDate)
    startModelDate$year <- startModelDate$year-predictionPeriodYears
    startModelDate <- as.Date(startModelDate)
    data <- results[results$date <= endPredictDate &
                      results$date > startModelDate,]
    # startIndex: first index that is not used to train the models
    startIndex <- max(which(data$date<startPredictDate))+1
    dataIds <- 1:(startIndex-1)
    fixId <- startIndex:nrow(data)
    if(NationalTeams==T){
    fixId<-fixId[which(data[fixId,"impScore"]!=1)]
    }
    # Reset startPredictDate for the next iteration
    startPredictDate <- as.Date(endPredictDate)+1
    if (length(fixId)!=0){
      print(length(fixId))
    data$days.passed <- as.numeric(difftime(max(
      data$date[fixId]),data$date,units="days"))
    appendSummary <- data[fixId,]
    appendSummary$date <- as.Date(appendSummary$date, format = "%Y-%m-%d")
    appendSummary[,"outcome"] <- data[fixId,"result"]
    ##################################################
    for(method in methods){
      print(method)
      print(system.time(
        for(j in 1:nbHalfPeriods){
          halfPeriod <- halfPeriods[j]
          data$timeEffect<-exp(-data$days.passed*log(2)/halfPeriod)
          data$weights<-data$timeEffect*data$impScore
          if(method %in% c("BTDgd","BTTgd","TMgd")){
            data$weights<-data$weights*data$gdweights
          }
          # Calculate the ML estimates
          nb.parameters<-ifelse(method=="KN1", nb.teams+2, 
                                ifelse(method=="P2",nb.teams*2-2+2,
                                       ifelse(method=="KN2",nb.teams*2-2+3,nb.teams+1 )))
          init <- doubleListLookup(inits,method,halfPeriod,nb.parameters)
          initB<-rep(0, nb.parameters)
          MLEfunction<-get(paste0("L",method)) 
          tryCatch({MLE <- optim(init,MLEfunction,gr = NULL, data=data[dataIds,], method="BFGS")
          }, error=function(e){tryCatch({
                    MLE <- optim(initB,MLEfunction,gr = NULL, data=data[dataIds,], method="BFGS")
          }, error=function(e){tryCatch({
                    MLE <- optim(init,MLEfunction,gr = NULL, data=data[dataIds,], method="CG")  }, 
             error=function(e){
                    MLE <- optim(initB,MLEfunction,gr = NULL, data=data[dataIds,], method="CG")  
                       })
            })
          })
          inits[[method]][[as.character(halfPeriod)]] <- MLE$par
          
          if (method %in% c("BTD","BTDgd")){
            strengths<-c(MLE$par[1:(nb.teams-1)],0)
            strengths<-strengths-mean(strengths)
            D <- MLE$par[nb.teams]
            H <- MLE$par[nb.teams+1]*(!data$neutral[fixId])
            diff<-strengths[data$homeId[fixId]]+H-strengths[data$visitId[fixId]]
            probH <- 1/(1+exp(D-diff/2)+exp(-diff))
            probA <- 1/(1+exp(D+diff/2)+exp(diff))
            probD <- 1-probH-probA
          }
          if (method %in% c("BTT", "BTTgd")){
            strengths<-c(MLE$par[1:(nb.teams-1)],0)
            strengths<-strengths-mean(strengths)
            D <- exp(MLE$par[nb.teams])
            H <- MLE$par[nb.teams+1]*(!data$neutral[fixId])
            diff<-strengths[data$homeId[fixId]]+H-strengths[data$visitId[fixId]]
            probH <- 1/(1+exp(D-diff))
            probA <- 1/(1+exp(D+diff))
            probD<-1-probH-probA
          }
          if (method %in% c("TM","TMgd")){
            strengths<-c(MLE$par[1:(nb.teams-1)],0)
            strengths<-strengths-mean(strengths)
            D <- exp(MLE$par[nb.teams])
            H <- MLE$par[nb.teams+1]*(!data$neutral[fixId])
            diff<-strengths[data$homeId[fixId]]+H-strengths[data$visitId[fixId]]
            probH <- pnorm(q=-D, mean=-diff, sd=1)
            probA <- pnorm(q=-D, mean=diff, sd=1)
            probD<-1-probH-probA
          }
          if (method %in% c("P1","KN1")){
            strengths<-c(MLE$par[1:(nb.teams-1)],0)
            strengths<-strengths-mean(strengths)
            C <- MLE$par[nb.teams]
            H <- MLE$par[nb.teams+1]*(!data$neutral[fixId])
            lambdasHome <- exp(C+H+strengths[data$homeId[fixId]]-strengths[data$visitId[fixId]])
            lambdasAway <- exp(C+strengths[data$visitId[fixId]]-strengths[data$homeId[fixId]])
            probH <- pskellam(-1,lambdasAway,lambdasHome)
            probD <- dskellam(0,lambdasAway,lambdasHome)
            probA <- 1-probH-probD
          }
          if (method %in% c("P2","KN2")){
            betasAttack<-c(MLE$par[1:(nb.teams-1)],0)
            betasDefense<-c(MLE$par[nb.teams:(2*nb.teams-2)],0)
            meanbetasAttack<-mean(betasAttack)
            meanbetasDefense<-mean(betasDefense)
            betasAttack<-betasAttack-meanbetasAttack
            betasDefense<-betasDefense-meanbetasDefense
            C <- MLE$par[nb.teams*2-1]+meanbetasAttack-meanbetasDefense
            H <- MLE$par[nb.teams*2]*(!data$neutral[fixId])
            lambdasHome <- exp(C+H+betasAttack[data$homeId[fixId]]-betasDefense[data$visitId[fixId]])
            lambdasAway <- exp(C+betasAttack[data$visitId[fixId]]-betasDefense[data$homeId[fixId]])
            probH <- pskellam(-1,lambdasAway,lambdasHome)
            probD <- dskellam(0,lambdasAway,lambdasHome)
            probA <- 1-probH-probD
          }
          outcomeProbs <- cbind(probH,probD,probA)
          outcomeProb <- outcomeProbs[1:nrow(outcomeProbs)+
                                        (1*as.numeric(data[fixId,"result"]=="D")+
                                           2*as.numeric(data[fixId,"result"]=="A")
                                        )*nrow(outcomeProbs)]
          appendSummary[[paste0("prob","_",method,"_",halfPeriod)]] <- outcomeProb
          appendSummary[[paste0("prob","_",method,"_",halfPeriod,"_H")]] <- outcomeProbs[,1]
          appendSummary[[paste0("prob","_",method,"_",halfPeriod,"_D")]] <- outcomeProbs[,2]
        } ))
    }  
    ###########################################################
    # Add all combined calculations to the summary data frame #
    ###########################################################
    summary <- data.frame(rbind(summary,appendSummary))
    rownames(summary) <- 1:nrow(summary)
    # Save each saveEachLoops iterations
    save(summary, file = targetSaveFile)
    }    
  }
  save(summary, file = targetSaveFile)
}

startPredictionDate<-ifelse(NationalTeams, "2008-01-01","2008-08-01")
predictionPeriodYears<-ifelse(NationalTeams, 8,2)
halfPeriods<-ifelse(NationalTeams, (1:12)*365/2, (1:24)*30)
targetSaveFile=paste0("predictions",competition,".Rdata")

computePredictions(startPredictionDate,
                   predictionPeriodYears,
                   halfPeriods, 
                   targetSaveFile, 
                   continueAnalysis=F,
                   NationalTeams,
                   methods=c("BTD","BTDgd","BTT","BTTgd","TM","TMgd","P1","P2","KN1","KN2"))




