################################
# Section 1: base ML functions #
################################

LBTD <- function(parameters, data){
  # Split up parameters
  n<-length(parameters)
  strengths<-c(parameters[1:(n-2)],0)
  D <- parameters[n-1]
  H <- parameters[n]*(as.numeric(!data$neutral))
  diff<-strengths[data$homeId]+H-strengths[data$visitId]
  # Calculate the outcome probabilities
  probH <- 1/(1+exp(D-diff/2)+exp(-diff))
  probA <- 1/(1+exp(D+diff/2)+exp(diff))
  probD<-1-probH-probA
  # Calculate negative log likelihood 
  out <- -sum(log(probH^((data$result=="H")*data$weights) *
                    probD^((data$result=="D")*data$weights) *
                    probA^((data$result=="A")*data$weights)))
  out
}
LBTDgd<-LBTD
LBTT <- function(parameters, data){
  # Split up parameters
  n<-length(parameters)
  strengths<-c(parameters[1:(n-2)],0)
  D <- exp(parameters[n-1]) # D has to be positive!
  H <- parameters[n]*(as.numeric(!data$neutral))
  diff<-strengths[data$homeId]+H-strengths[data$visitId]
  # Calculate the outcome probabilities
  probH <- 1/(1+exp(D-diff))
  probA <- 1/(1+exp(D+diff))
  probD<-1-probH-probA
  # Calculate negative log likelihood 
  out <- -sum(log(probH^((data$result=="H")*data$weights) *
                    probD^((data$result=="D")*data$weights) *
                    probA^((data$result=="A")*data$weights)))
  out
}
LBTTgd<-LBTT
LTM <- function(parameters, data){
  # Split up parameters
  n<-length(parameters)
  strengths<-c(parameters[1:(n-2)],0)
  D <- exp(parameters[n-1]) # D has to be positive!
  H <- parameters[n]*(as.numeric(!data$neutral))
  diff<-strengths[data$homeId]+H-strengths[data$visitId]
  # Calculate the outcome probabilities
  probH <- pmax(pnorm(q=-D, mean=-diff, sd=1), 1/10^300)
  probA <- pmax(pnorm(q=-D, mean=diff, sd=1), 1/10^300)
  probD<-pmax(1-probH-probA, 1/10^300)
  # Calculate negative log likelihood 
  out <- -sum(log(probH^((data$result=="H")*data$weights) *
                    probD^((data$result=="D")*data$weights) *
                    probA^((data$result=="A")*data$weights)))
  out
  
}
LTMgd<-LTM
LP1 <- function(parameters, data){
  # Split up parameters
  n <- length(parameters)
  strengths<-c(parameters[1:(n-2)],0)
  C <- parameters[n-1]
  H <- parameters[n]*(!data$neutral)
  # Calculate lambdas for all matches and both teams
  loglambdasHome <- C+H+strengths[data$homeId]-strengths[data$visitId]
  loglambdasAway <- C+strengths[data$visitId]-strengths[data$homeId]
  # Calculate negative log likelihood 
  out <- -sum((loglambdasHome*data$home_score-exp(loglambdasHome)
              +loglambdasAway*data$away_score-exp(loglambdasAway))*(data$weights))
  out
}
LP2 <- function(parameters, data){
  # Split up parameters
  n <- length(parameters)
  nb.teams<-(n)/2
  attstrengths<-c(parameters[1:(nb.teams-1)],0)
  defstrengths<-c(parameters[nb.teams:(2*nb.teams-2)],0)
  C <- parameters[n-1]
  H <- parameters[n]*(as.numeric(!data$neutral))
  # Calculate lambdas for all matches and both teams
  loglambdasHome <- C+H+attstrengths[data$homeId]-defstrengths[data$visitId]
  loglambdasAway <- C+attstrengths[data$visitId]-defstrengths[data$homeId]
  # Calculate negative log likelihood 
  out <- -sum((loglambdasHome*data$home_score-exp(loglambdasHome)
               +loglambdasAway*data$away_score-exp(loglambdasAway))*(data$weights))
  out
}
LKN1 <- function(parameters, data){
  # Split up parameters
  n <- length(parameters)
  strengths<-c(parameters[1:(n-3)],0)
  C<-parameters[n-2]
  H<-parameters[n-1]*(as.numeric(!data$neutral))
  # Calculate lambdas for all matches and both teams
  loglambdasHome <- C+H+strengths[data$homeId]-strengths[data$visitId]
  loglambdasAway <- C+strengths[data$visitId]-strengths[data$homeId]
  lambdasHome<-exp(loglambdasHome)
  lambdasAway<-exp(loglambdasAway)
  lambdas3<-exp(parameters[n])
  # Calculate negative log likelihood 
  logchances<- -(lambdasHome+lambdasAway+lambdas3)+loglambdasHome*data$home_score+loglambdasAway*data$away_score
  lambdaratio<-lambdas3/(lambdasHome*lambdasAway) 
  som<-1+data$kn1*lambdaratio+data$kn2*lambdaratio^2+data$kn3*lambdaratio^3+data$kn4*lambdaratio^4+data$kn5*lambdaratio^5
  out <- -sum((logchances+log(som))*(data$weights))
  out
}
LKN2 <- function(parameters, data){
  # Split up parameters
  n <- length(parameters)
  nb.teams<-(n-1)/2
  attstrengths<-c(parameters[1:(nb.teams-1)],0)
  defstrengths<-c(parameters[nb.teams:(2*nb.teams-2)],0)
  C<-parameters[n-2]
  H<-parameters[n-1]*(as.numeric(!data$neutral))
  # Calculate lambdas for all matches and both teams
  loglambdasHome <- C+H+attstrengths[data$homeId]-defstrengths[data$visitId]
  loglambdasAway <- C+attstrengths[data$visitId]-defstrengths[data$homeId]
  lambdasHome<-exp(loglambdasHome)
  lambdasAway<-exp(loglambdasAway)
  lambdas3<-exp(parameters[n])
  # Calculate negative log likelihood
  logchances<- -(lambdasHome+lambdasAway+lambdas3)+loglambdasHome*data$home_score+loglambdasAway*data$away_score
  lambdaratio<-lambdas3/(lambdasHome*lambdasAway) 
  som<-1+data$kn1*lambdaratio+data$kn2*lambdaratio^2+data$kn3*lambdaratio^3+data$kn4*lambdaratio^4+data$kn5*lambdaratio^5
  out <- -sum((logchances+log(som))*(data$weights))
  out
}

